from __future__ import print_function, unicode_literals
from ..logger import Logger
from halo import Halo
import sys
import boto3
import time

logger = Logger()

cloudformation = boto3.client('cloudformation')
cloudfront = boto3.client('cloudfront')


def update_stack(stack_name, version):
    params = update_template_params(stack_name, version)
    if params:
        logger.info(
            "Effettuo update template cloudformation {}".format(stack_name))
        response = cloudformation.update_stack(
            StackName=stack_name,
            TemplateBody=get_template_from_stack_name(stack_name),
            Parameters=params
        )
    else:
        logger.error(
            "Nessun parametro trovato sullo stack {}".format(stack_name))
        sys.exit(-1)


def update_template_params(stack_name, version):
    new_parameter = []
    ReleaseVersionFound = False
    current_parameters = get_parameters_from_stack_name(stack_name)

    for param in current_parameters:
        if param["ParameterKey"] == "ReleaseVersion":
            new_parameter.append(
                {"ParameterKey": param["ParameterKey"], "ParameterValue": version})
            ReleaseVersionFound = True
        else:
            new_parameter.append(
                {"ParameterKey": param["ParameterKey"], "UsePreviousValue": True})

    if ReleaseVersionFound:
        return new_parameter
    else:
        return None


def get_parameters_from_stack_name(stack_name):
    return cloudformation.describe_stacks(StackName=stack_name)["Stacks"][0]["Parameters"]


def get_template_from_stack_name(stack_name):
    return cloudformation.get_template(StackName=stack_name)["TemplateBody"]


def get_stacks_name(repo):
    stacks_name = []
    base_stacks_name = ["ecs-task-{}-production".format(repo),
                        "ecs-task-{}-vpc-production".format(repo),
                        "ecs-job-{}-production".format(repo),
                        "batch-job-{}-production".format(repo)]

    for stack_name in base_stacks_name:
        if stack_exists(stack_name):
            stacks_name.append(stack_name)

    return stacks_name


def stack_exists(stack_name):
    try:
        cloudformation.describe_stacks(StackName=stack_name)
        return True
    except:
        return False


def get_stack_status(stack_name):
    return cloudformation.describe_stacks(StackName=stack_name)["Stacks"][0]["StackStatus"]


def get_artifacts_from_s3(bucket_name, prefix):
    with Halo(text='Loading releases...', spinner='dots', color='magenta'):
        s3 = boto3.resource('s3')
        bucket_artifacts = s3.Bucket(bucket_name)
        artifacts = []

        for bucket_object in bucket_artifacts.objects.filter(Prefix=prefix):
            artifacts.append(bucket_object.key.replace(prefix, ''))

    return artifacts


def is_cloudfront_distribution(repo):
    with Halo(text='Loading releases...', spinner='dots', color='magenta'):
        distributions = cloudfront.list_distributions()[
            "DistributionList"]["Items"]
        for distribution in distributions:
            if "prima-prod-{}.s3.amazonaws.com".format(repo) == distribution["Origins"]["Items"][0]["DomainName"]:
                return True
            else:
                continue
    return False


def get_tag_from_s3_artifact(bucket_name, prefix, artifact):
    s3 = boto3.client('s3')

    return s3.get_object_tagging(
        Bucket=bucket_name,
        Key="{}{}".format(prefix, artifact)
    )["TagSet"]


def stack_update_completed(stack_name):
    with Halo(text='Rollback in progress...', spinner='dots', color='magenta'):
        for i in range(0, 60):
            stack_status = get_stack_status(stack_name)
            if stack_status == "UPDATE_COMPLETE":
                return True
            elif "FAILED" in stack_status:
                break
            else:
                time.sleep(10)
    return False


def update_stacks(stacks_name, version):
    for stack_name in stacks_name:
        update_stack(stack_name, version)

    for stack_name in stacks_name:
        if not stack_update_completed(stack_name):
            logger.error(
                "Errore durante l'update dello stack CloudFormation {}. Contattare i DevOps".format(stack_name))
            sys.exit(-1)



from __future__ import print_function, unicode_literals
from .singleton import Singleton
import os
import yaml


class Config(metaclass=Singleton):
    _config = {}

    def __init__(self):
        pass

    def load(self):
        with open(os.path.join(os.environ['HOME'], '.suite_py/config.yml')) as configfile:
            self._config = yaml.safe_load(configfile)

        self._config['user']['projects_home'] = os.path.join(
            os.environ['HOME'], self._config['user']['projects_home'])
        return self._config



from __future__ import print_function, unicode_literals
from termcolor import colored
from .singleton import Singleton


class Logger(metaclass=Singleton):

    def __init__(self):
        pass

    def info(self, message):
        print(colored(message, 'green'))

    def warning(self, message):
        print(colored(message, 'yellow'))

    def error(self, message):
        print(colored(message, 'red'))



from __future__ import print_function, unicode_literals
from ..tokens import Tokens
from ..singleton import Singleton
from github import Github
from github.GithubException import GithubException, UnknownObjectException
import os

tokens = Tokens()


class GithubHandler(metaclass=Singleton):
    _client = None
    _organization = 'primait'

    def __init__(self):
        self._client = Github(tokens.github)

    def get_repo(self, repo_name):
        return self._client.get_repo('{}/{}'.format(self._organization, repo_name))

    def get_organization(self):
        return self._client.get_organization(self._organization)

    def get_user(self):
        return self._client.get_user()

    def create_pr(self, repo, branch, title, body=''):
        return self.get_repo(repo).create_pull(title=title, head=branch, base='master', body=body)

    def get_pr(self, repo, pr_number):
        return self.get_repo(repo).get_pull(pr_number)

    def get_branch_from_pr(self, repo, pr_number):
        repo = self.get_repo(repo)
        return repo.get_branch(repo.get_pull(pr_number).head.ref)

    def get_team_members(self, team_name=""):
        return self.get_organization().get_team_by_slug(team_name).get_members()

    def get_all_users(self):
        return self.get_organization().get_members()

    def get_list_pr(self, repo):
        pulls = self.get_repo(repo).get_pulls(
            state='open', sort='created', base='master')
        return pulls

    def get_pr_from_branch(self, repo, branch):
        return self.get_repo(repo).get_pulls(
            head="{}:{}".format("primait", branch))

    def get_link_from_pr(self, repo, pr_number):
        return 'https://github.com/primait/{}/pull/{}'.format(repo, str(pr_number))

    def get_commits_since_release(self, repo, release):
        release_commit = repo.get_commit(release.tag_name)
        commits = []
        for c in repo.get_commits():
            if c == release_commit:
                break
            else:
                commits.append(c)
        return commits

    def get_latest_release_if_exists(self, repo):
        try:
            return repo.get_latest_release()
        except:
            return None

    def user_is_admin(self, repo):
        return self.get_repo(repo).permissions.admin

    def get_release_if_exists(self, repo, release):
        try:
            return repo.get_release(release)
        except:
            return None



from __future__ import print_function, unicode_literals
from halo import Halo
from ..lib.handler.youtrack_handler import YoutrackHandler
from ..lib.handler.github_handler import GithubHandler
from ..lib.handler import git_handler as git
from ..lib.handler import drone_handler as drone
from ..lib.handler import prompt_utils
from ..lib.logger import Logger
from ..lib.qainit import get_qa_projects, qainit_deploy
from ..lib.config import Config

import subprocess
import sys
import json

youtrack = YoutrackHandler()
github = GithubHandler()
logger = Logger()
config = Config().load()


def entrypoint(args):
    qa_projects = get_qa_projects()
    if args.project not in qa_projects:
        logger.error("Il progetto {} non fa parte degli ambienti di QA, niente da fare".format(args.project))
        sys.exit(0)

    with Halo(text='Updating prima-twig...', spinner='dots', color='magenta'):
        try:
            subprocess.run(['gem', 'update', 'prima-twig'],
                check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except:
            logger.error("Errore durante l'aggiornamento di prima-twig, per favore esegui manualmente `gem update prima-twig`")
            sys.exit(-1)

    branch_name = git.current_branch_name()
    youtrack_id = youtrack.get_card_from_name(branch_name)

    if youtrack_id:
        if prompt_utils.ask_confirm("Ci sono altri progetti da aggiungere al QA che fanno riferimento alla card {}?".format(youtrack_id), default=False):
            twig_arg = multi_repo_qa(youtrack_id, qa_projects)
        else:
            twig_arg = mono_repo_qa(args.project, branch_name)
    else:
        logger.warning("Non sono riuscito a trovare una issue YouTrack collegata al branch, non e' possibile collegare automaticamente altri progetti a questo QA")
        logger.warning("Se vuoi creare il QA con diversi branch su diversi progetti per favore interrompi il comando ed utilizza twig")
        twig_arg = mono_repo_qa(args.project, branch_name)

    qainit_deploy(twig_arg)

    if youtrack_id:
        logger.info("Aggiorno la card su youtrack...")
        youtrack.update_state(youtrack_id, config["youtrack"]["test_state"])

    drone_url = drone.get_last_build_url('qainit')
    if drone_url:
        logger.info("Puoi seguire la creazione del QA su {}".format(drone_url))

    logger.info("Configurazione del QA effettuata con successo!")


def multi_repo_qa(youtrack_id, qa_projects):
    logger.info("Cerco gli altri branch su github...")
    repos = list(map(lambda project: github.get_repo(project), qa_projects))

    selected_repos = []
    for repo in repos:
        print('.', end='', flush=True)
        for branch in repo.get_branches():
            if youtrack_id in branch.name:
                selected_repos.append((repo.name, branch))

    print('.')
    logger.info("Progetti trovati:")
    for elem in selected_repos:
        logger.info("{} --> {}".format(elem[0], elem[1].name))

    return twigify(selected_repos)

def mono_repo_qa(project, branch_name):
    logger.info("Creo il qa con il progetto {} e branch {}".format(project, branch_name))
    gh_branch = github.get_repo(project).get_branch(branch_name)
    selected_repos = [(project, gh_branch)]

    return twigify(selected_repos)

def twigify(projects_list):
    projects_dict = {}
    for tuple in projects_list:
        projects_dict[tuple[0]] = {"name": tuple[1].name,
                                   "revision": tuple[1].commit.sha[:15],
                                   "committer": tuple[1].commit.committer.email or "",
                                   "default_branch": False}

    return json.dumps(json.dumps(projects_dict)) # necessario per fare l'escape corretto

    # development only
    # return json.dumps(projects_dict) chiamando twig da file non bisogna dumpare due volte



from __future__ import print_function, unicode_literals
from subprocess import CalledProcessError
from ..logger import Logger
from ..config import Config
import os
import subprocess
import sys
import re
import semver
from halo import Halo

logger = Logger()
config = Config().load()


def _get_path(repo):
    return os.path.join(config['user']['projects_home'], repo)


def clone(repo):
    subprocess.run(['git', 'clone', 'git@github.com:primait/{}.git'.format(repo)],
                   cwd=config['user']['projects_home'], check=True)


def checkout(repo, branch, new=False):
    try:
        if new:
            subprocess.run(['git', 'checkout', '-b', branch],
                           cwd=_get_path(repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        else:
            subprocess.run(['git', 'checkout', branch],
                           cwd=_get_path(repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            pull(repo, rebase=True)
    except (CalledProcessError) as e:
        if (e.returncode == 128):
            checkout(repo, branch)  # caso checkout -b e branch gia' esistente
        elif (e.returncode == 1):
            # caso checkout normale e branch non esistente
            checkout(repo, branch, new=True)
        else:
            logger.error("Errore eseguendo il comando: {}".format(e))
            sys.exit(-1)


def commit(repo, commit_message='', dummy=False):
    try:
        if dummy:
            subprocess.run(['git', 'commit',  '--allow-empty', '-m',
                            commit_message], cwd=_get_path(repo), check=True)
        else:
            subprocess.run(['git', 'commit',  '-am', commit_message],
                           cwd=_get_path(repo), check=True)
    except (CalledProcessError) as e:
        logger.error("Errore eseguendo il comando: {}".format(e))
        sys.exit(-1)


def push(repo, branch, remote='origin'):
    try:
        subprocess.run(['git', 'push', remote, branch],
                       cwd=_get_path(repo), check=True)
    except (CalledProcessError) as e:
        logger.error("Errore eseguendo il comando: {}".format(e))
        sys.exit(-1)


def pull(repo, rebase=False):
    with Halo(text='Pulling...', spinner='dots', color='magenta'):
        try:
            if rebase:
                subprocess.run(['git', 'pull', "--rebase"],
                               cwd=_get_path(repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            else:
                subprocess.run(['git', 'pull'],
                               cwd=_get_path(repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except (CalledProcessError) as e:
            logger.error("Errore eseguendo il comando: {}".format(e))
            sys.exit(-1)


def fetch(repo, remote='origin'):
    with Halo(text='Fetching...', spinner='dots', color='magenta'):
        try:
            subprocess.run(['git', 'fetch', '--quiet'], cwd=_get_path(
                repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            subprocess.run(['git', 'fetch', '-p', remote, '--quiet'],
                           cwd=_get_path(repo), check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        except (CalledProcessError) as e:
            logger.error("Errore eseguendo il comando: {}".format(e))
            sys.exit(-1)


def sync(repo):
    fetch(repo)
    checkout(repo, 'master')
    pull(repo)


def check_repo_cloned(repo):
    if repo not in os.listdir(config['user']['projects_home']):
        logger.warning(
            "Il progetto non e' presente nella tua home dei progetti, lo clono adesso..")
        clone(repo)


def delete_remote_branch(repo, branch):
    try:
        push(repo, ":{}".format(branch))
        return True
    except BaseException:
        return False


def get_git_username():
    try:
        output = subprocess.run(['git', 'config', 'user.name'],
                                check=True, stdout=subprocess.PIPE).stdout.decode('utf-8')
        output = re.sub("[^A-Za-z]", "", output)
        return re.sub("\n", "", output)
    except (CalledProcessError):
        return None


def get_last_tag(project):
    sync(project)
    return subprocess.check_output(["git", "describe", "--abbrev=0", "--tags"]).decode('utf-8').strip()


def search_remote_branch(repo, regex):
    try:
        output = subprocess.run(['git', 'branch', '-r', '--list', regex],
                                cwd=_get_path(repo), check=True, stdout=subprocess.PIPE)

        return output.stdout.decode("utf-8").strip('\n').strip().strip('origin/')
    except (CalledProcessError):
        return ""


def remote_branch_exists(repo, branch):
    try:
        subprocess.run(['git', 'show-ref', '--quiet', 'refs/remotes/origin/{}'.format(branch)],
                       cwd=_get_path(repo), check=True)
        return True
    except (CalledProcessError):
        return False


def local_branch_exists(repo, branch):
    try:
        subprocess.run(['git', 'show-ref', '--quiet', 'refs/heads/{}'.format(branch)],
                       cwd=_get_path(repo), check=True)
        return True
    except (CalledProcessError):
        return False


def current_branch_name():
    try:
        output = subprocess.run(['git', 'rev-parse', '--abbrev-ref', 'HEAD'],
                                stdout=subprocess.PIPE).stdout.decode('utf-8')
        return re.sub("\n", "", output)
    except (CalledProcessError) as e:
        logger.error("Errore eseguendo il comando: {}".format(e))
        sys.exit(-1)


def is_dirty(repo):
    try:
        subprocess.run(['git', 'diff', '--quiet'], check=True)
        subprocess.run(
            ['git', 'diff', 'HEAD', '--name-only', '--quiet'], check=True)
        return False
    except (CalledProcessError) as e:
        logger.error(
            "Hai modifiche non committate su {} oppure non è un repository!".format(repo))
        sys.exit(-1)


def files_changed_between_commits(from_commit, to_commit):
    p = subprocess.Popen(
        ["git", "diff", from_commit, to_commit, "--name-only"], stdout=subprocess.PIPE)
    result = p.communicate()[0]
    return result.decode('utf-8').splitlines()


def migrations_found(files_changed):
    for file in files_changed:
        if 'migration' in file.lower():
            return True
    return False


def get_last_tag_number(tags):
    for tag in tags:
        if is_semver(tag.name):
            return tag.name
        else:
            continue
    logger.warning("Nessun tag trovato, sto per pubblicare tag 0.1.0")
    return "0.1.0"


def is_semver(tag):
    try:
        return semver.parse(tag)
    except:
        return None



from __future__ import print_function, unicode_literals
from ..lib.handler.github_handler import GithubHandler
from ..lib.handler.slack_handler import SlackHandler
from ..lib.handler import prompt_utils
from ..lib.handler import aws_handler as aws
from ..lib.handler import drone_handler as drone
from ..lib.logger import Logger
from distutils.version import StrictVersion
from halo import Halo
import sys
import re

github = GithubHandler()
logger = Logger()
slack = SlackHandler()


def entrypoint(args):
    bucket_name = "prima-artifacts-encrypted"
    repo = github.get_repo(args.project)

    if args.project == "prima":
        stacks_name = ["ecs-task-web-vpc-production", "ecs-task-consumer-vpc-production",
                       "ecs-task-cron-vpc-production", "batch-job-php-production"]
        prefix = "prima/"

        artifacts = aws.get_artifacts_from_s3(
            bucket_name, prefix)
        versions, prima_version_mapping = get_versions_from_artifacts(
            args.project, bucket_name, artifacts)

        version = ask_version(repo, versions)
        rollback_stacks(args.project, stacks_name,
                        prima_version_mapping[version])

    elif aws.is_cloudfront_distribution(args.project):
        with Halo(text='Loading releases...', spinner='dots', color='magenta'):
            versions = drone.get_tag_from_builds(args.project)
        version = ask_version(repo, versions)
        build = drone.get_build_from_tag(args.project, version)
        drone.launch_build(args.project, build)
        logger.info("Build rilanciata, controlla lo stato su drone")
        post_rollback_complete_on_slack(args.project, version)

    else:
        stacks_name = aws.get_stacks_name(args.project)
        prefix = "microservices/{}/".format(args.project)

        if len(stacks_name) > 0:
            artifacts = aws.get_artifacts_from_s3(
                bucket_name, prefix)
            versions, _ = get_versions_from_artifacts(
                args.project, bucket_name, artifacts)

            if len(versions) > 0:
                version = ask_version(repo, versions)
                rollback_stacks(args.project, stacks_name, version)
            else:
                logger.error(
                    "Nessuna release trovata. Impossibile procedere con il rollback.")
                sys.exit(-1)

        else:
            logger.error(
                "Nessuno stack trovato. Impossibile procedere con il rollback.")


def ask_version(repo, choiches):
    version = prompt_utils.ask_choices("Seleziona release: ", choiches)
    release = github.get_release_if_exists(repo, version)
    logger.info(
        "\nDescrizione della release selezionata:\n{}\n".format(release.body))
    if not prompt_utils.ask_confirm("Vuoi continuare con il rollback?"):
        sys.exit(-1)
    return version


def get_versions_from_artifacts(project, bucket_name, artifacts):
    with Halo(text='Loading releases...', spinner='dots', color='magenta'):

        versions = []
        prima_version_mapping = {}

        for artifact in artifacts:

            if project == "prima":

                tags_object = aws.get_tag_from_s3_artifact(
                    bucket_name, "prima/", artifact)

                for tag_object in tags_object:
                    if tag_object["Key"] == "ReleaseVersion":
                        versions.append(tag_object["Value"])
                        prima_version_mapping[tag_object["Value"]] = artifact.replace(
                            ".tar.gz", "")

            else:
                if re.match("[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}-production.tar.gz", artifact):
                    versions.append(artifact.replace("-production.tar.gz", ""))

        versions.sort(key=StrictVersion, reverse=True)
    return versions, prima_version_mapping


def rollback_stacks(project, stacks_name, version):
    aws.update_stacks(stacks_name, version)
    post_rollback_complete_on_slack(project, version)
    logger.info(
        "Rollback completato con successo. Versione in produzione: {}".format(version))


def post_rollback_complete_on_slack(project, version):
    try:
        slack.post("#deploy", "ho effettuato il rollback di {}. Versione in produzione: {}".format(
            project, version))
    except:
        logger.warning("Non sono riuscito a inviare il messaggo su Slack")



from ..tokens import Tokens
from distutils.version import StrictVersion
from halo import Halo
import requests
import time

tokens = Tokens()
baseurl = "https://drone-1.prima.it"


def get_last_build_url(repo):
    with Halo(text='Contacting drone...', spinner='dots', color='magenta'):
        # necessario per far comparire la build che abbiamo appena pushato
        time.sleep(2)
        try:
            resp = requests.get("{}/api/repos/primait/{}/builds".format(baseurl, repo),
                                headers={'Authorization': 'Bearer {}'.format(tokens.drone)}).json()

            return '{}/primait/{}/{}'.format(baseurl, repo, resp[0]['number'])
        except (Exception) as e:
            return ""


def get_pr_build_url(repo, commit_sha):
    with Halo(text='Contacting drone...', spinner='dots', color='magenta'):
        # necessario per far comparire la build che abbiamo appena pushato
        time.sleep(2)
        try:
            resp = requests.get("{}/api/repos/primait/{}/builds?per_page=100".format(baseurl, repo),
                                headers={'Authorization': 'Bearer {}'.format(tokens.drone)}).json()
            build_number = list(filter(lambda build: build["after"] == commit_sha, resp))[
                0]['number']
            return '{}/primait/{}/{}'.format(baseurl, repo, build_number)
        except (Exception) as e:
            return ""


def get_tag_from_builds(repo):
    tags = []
    builds = requests.get("{}/api/repos/primait/{}/builds?per_page=100".format(baseurl, repo),
                          headers={'Authorization': 'Bearer {}'.format(tokens.drone)}).json()

    for build in builds:
        if build["event"] == "tag":
            tags.append(build["ref"].replace('refs/tags/', ''))

    tags = list(dict.fromkeys(tags))
    tags.sort(key=StrictVersion, reverse=True)
    return tags


def get_build_from_tag(repo, tag):
    builds = requests.get("{}/api/repos/primait/{}/builds?per_page=100".format(baseurl, repo),
                          headers={'Authorization': 'Bearer {}'.format(tokens.drone)}).json()

    for build in builds:
        if build["event"] == "tag":
            if build["ref"].replace('refs/tags/', '') == tag:
                return build["number"]
    return None


def launch_build(repo, build):
    return requests.post("{}/api/repos/primait/{}/builds/{}".format(baseurl, repo, build),
                         headers={'Authorization': 'Bearer {}'.format(tokens.drone)}).json()



from __future__ import print_function, unicode_literals
from ..singleton import Singleton
from ..handler import git_handler as git
import os
import requests
import sys


class CaptainHook(metaclass=Singleton):
    _baseurl = None

    def __init__(self):
        self._baseurl = "http://captainhook-internal.prima.it"

    def lock_project(self, project):
        data = {'project': project, 'status': "locked", "user": git.get_git_username()}
        return self.send_post_request("/projects/manage-lock", data)

    def unlock_project(self, project):
        data = {'project': project, 'status': "unlocked", "user": git.get_git_username()}
        return self.send_post_request("/projects/manage-lock", data)

    def status(self, project):
        return self.send_get_request("/projects/check?project={}".format(project))

    def get_users_list(self):
        return self.send_get_request("/users/all")

    def send_post_request(self, endpoint, data):
        return requests.post("{}{}".format(self._baseurl, endpoint), data=data)

    def send_get_request(self, endpoint):
        return requests.get("{}{}".format(self._baseurl, endpoint))



from __future__ import print_function, unicode_literals
from ..lib.handler.youtrack_handler import YoutrackHandler
from ..lib.handler.github_handler import GithubHandler
from ..lib.logger import Logger
from ..lib.handler import git_handler as git
from ..lib.handler.captainhook_handler import CaptainHook
from ..lib.handler.slack_handler import SlackHandler
import sys
import readline

youtrack = YoutrackHandler()
github = GithubHandler()
logger = Logger()
slack = SlackHandler()

try:
    captainhook = CaptainHook()
    users = captainhook.get_users_list().json()
except:
    logger.error(
        "Non riesco ad ottenere la lista degli utenti da captainhook.\
        Controlla di avere attiva la VPN. Se il problema persiste aggiungi i reviewer dal sito GitHub")
    sys.exit(-1)


def entrypoint(args):
    branch_name = git.current_branch_name()

    github_reviewers = []
    slack_reviewers = []

    pull = github.get_pr_from_branch(args.project, branch_name)

    if pull.totalCount:
        pr = pull[0]
        logger.info('Ho trovato la pull request numero {} per il branch {} sul repo {}'.format(
            pr.number, branch_name, args.project))
    else:
        logger.error(
            'Nessuna pull request aperta trovata per il branch {}'.format(branch_name))
        sys.exit(-1)

    youtrack_reviewers = ask_reviewer()

    for rev in youtrack_reviewers:
        for user in users:
            if user["youtrack"] == rev:
                github_reviewers.append(user["github"])
                slack_reviewers.append(user["slack"])

    pr.create_review_request(github_reviewers)
    logger.info('Aggiungo reviewers su GitHub')

    revs = ""
    for rev in slack_reviewers:
        revs += "<@{}> ".format(rev)

    slack.post('#review', '{}{}'.format(
        revs, pr.html_url))

    logger.info('Cito reviewers su Slack')

    youtrack_id = youtrack.get_card_from_name(pr.title)

    if youtrack_id:
        logger.info(
            'Sposto la card {} in review su youtrack e aggiungo i tag degli utenti'.format(youtrack_id))
        youtrack.update_state(youtrack_id, 'Review')
        for rev in youtrack_reviewers:
            try:
                youtrack.add_tag(youtrack_id, "review:{}".format(rev))
            except BaseException as e:
                logger.warning(
                    "Non sono riuscito ad aggiungere i tag di review: {}".format(e))
                sys.exit(-1)
    else:
        logger.warning(
            'Reviewers inseriti SOLO su GitHub. Nessuna card collegata o card nel nome del branch inesistente su YouTrack.')


def ask_reviewer():
    readline.set_completer(completer)
    readline.parse_and_bind("tab: complete")

    youtrack_reviewers = []

    youtrack_reviewers = list(
        input("Scegli i reviewers (nome.cognome - separati da spazio - premere TAB per autocomplete) > ").split())

    if len(youtrack_reviewers) < 1:
        logger.warning("Devi inserire almeno un reviewer")
        return ask_reviewer()
    else:
        return youtrack_reviewers


def completer(text, state):
    options = [x["youtrack"]
               for x in users if text.lower() in x["youtrack"].lower()]
    try:
        return options[state]
    except IndexError:
        return None



from __future__ import print_function, unicode_literals
from ..lib.handler.youtrack_handler import YoutrackHandler
from ..lib.handler.github_handler import GithubHandler
from ..lib.handler.captainhook_handler import CaptainHook
from ..lib.config import Config
from ..lib.logger import Logger
from ..lib.handler import prompt_utils
from ..lib.qainit import qainit_shutdown
from ..lib.handler import git_handler as git
from ..lib.handler import drone_handler as drone
import os
import sys
import re

youtrack = YoutrackHandler()
github = GithubHandler()
logger = Logger()
lock = CaptainHook()
config = Config().load()


def entrypoint(args):
    stop_if_master_locked(args.project)

    pr = select_pr(args.project)
    branch_name = pr.head.ref
    youtrack_id = youtrack.get_card_from_name(branch_name)

    check_migrations_merge(pr)

    logger.info("Eseguo il merge...")

    merge_status = pr.merge(
        commit_title="{} (#{})".format(pr.title, pr.number), commit_message='', merge_method='squash')

    if not merge_status.merged:
        logger.error("Si è verificato un errore durante il merge.")
        sys.exit(-1)

    drone_build = get_drone_build(args.project, merge_status.sha)

    if drone_build:
        logger.info(
            "Pull request mergiata su master! Puoi seguire lo stato della build su {}".format(drone_build))
    else:
        logger.info("Pull request mergiata su master!")

    git.fetch(args.project)
    if git.remote_branch_exists(args.project, branch_name):
        git.delete_remote_branch(args.project, branch_name)

    if prompt_utils.ask_confirm("Vuoi bloccare staging? (Necessario se bisogna testare su staging)".format(args.project), default=False):
        lock.lock_project(args.project)

    if youtrack_id:
        logger.info("Aggiorno lo stato della card su youtrack...")
        youtrack.update_state(youtrack_id, config["youtrack"]["merged_state"])
        logger.info("Card aggiornata")

        logger.info("Spengo il qa, se esiste")
        qainit_shutdown(youtrack_id)
    else:
        logger.warning(
            "Non sono riuscito a trovare una issue YouTrack nel nome del branch o la issue indicata non esiste.")
        logger.warning(
            "Nessuna card aggiornata su YouTrack e nessun QA spento in automatico")

    logger.info("Tutto fatto!")
    sys.exit()


def select_pr(project):
    if github.user_is_admin(project):
        logger.info("Ottengo le pull request da GitHub...")
        logger.warning(
            "Sei admin del repository, puoi fare il merge skippando i check (CI, review, ecc...)\nDa grandi poteri derivano grandi responsabilita'")
        prs = github.get_list_pr(project)
    else:
        logger.info("Ottengo le pull request con check verdi da GitHub...")
        prs = list(filter(lambda pr: pr.mergeable and pr.mergeable_state ==
                          "clean", github.get_list_pr(project)))

    if prs.totalCount > 0:
        choices = [{"name": "{} {}".format(
            i.number, i.title), "value": i} for i in prs]
        return prompt_utils.ask_choices('Seleziona PR: ', choices)
    else:
        logger.warning(
            "Non esistono pull request pronte per il merge, per favore controlla su https://github.com/primait/{}/pulls".format(project))
        sys.exit(-1)


def stop_if_master_locked(project):
    request = lock.status(project)
    if request.status_code != 200:
        logger.error("Impossibile determinare lo stato del lock su master.")
        sys.exit(-1)

    if request.json()["locked"]:
        logger.error(
            "Il progetto è lockato su master. Impossibile continuare.")
        sys.exit(-1)


def get_drone_build(project, sha):
    drone_build = drone.get_pr_build_url(
        project, sha)
    if drone_build != "":
        return drone_build
    return None


def check_migrations_merge(pr):
    files_changed = [x.filename for x in pr.get_files()]
    if git.migrations_found(files_changed):
        logger.warning("ATTENZIONE: migrations rilevate nel codice")
        if not prompt_utils.ask_confirm("Sicuro di voler continuare?"):
            sys.exit(-1)
